import pytest
from _pytest.pathlib import Path
import typing as T
import json
import sys
from .__version__ import __version__
import tqdm
import math


PYTEST_ARGS = ('-qq', '-p', 'pytest_cleanslate.reduce')
MODULE_LIST_ARG = '--module-list-from'
TEST_LIST_ARG = '--test-list-from'
RESULTS_ARG = '--results-to'


class ReducePlugin:
    def __init__(self, config: pytest.Config) -> None:
        if (modules_file := config.getoption(MODULE_LIST_ARG)):
            with modules_file.open('r') as f:
                self._modules = {Path(line.strip()).resolve() for line in f}
        else:
            self._modules = None

        if (tests_file := config.getoption(TEST_LIST_ARG)):
            with tests_file.open('r') as f:
                self._tests = set(line.strip() for line in f)
        else:
            self._tests = None

        self._results_file = config.getoption(RESULTS_ARG)
        self._collect = []
        self._run = []


    @pytest.hookimpl
    def pytest_ignore_collect(self, collection_path: Path, config: pytest.Config) -> T.Union[None, bool]:
        if self._modules is not None and collection_path.suffix == '.py':
            return collection_path.resolve() not in self._modules


    @pytest.hookimpl
    def pytest_collectreport(self, report: pytest.CollectReport) -> None:
        if self._results_file and report.nodeid.endswith('.py'):
            self._collect.append({
                'id': report.nodeid,
                'outcome': report.outcome,
                'result': [n.nodeid for n in report.result]
            })


    @pytest.hookimpl(tryfirst=True)
    def pytest_collection_modifyitems(self, items: T.List[pytest.Item], config: pytest.Config) -> T.Union[None, bool]:
        if self._tests is not None:
            selected = []
            deselected = []

            for item in items:
                if item.nodeid in self._tests:
                    selected.append(item)
                else:
                    deselected.append(item)

            if deselected:
                config.hook.pytest_deselected(items=deselected)
                items[:] = selected


    @pytest.hookimpl
    def pytest_runtest_logreport(self, report: pytest.TestReport) -> None:
        # to cut down on the log, we only save non-call run reports for failures
        if self._results_file and (report.outcome != 'passed' or report.when == 'call'):
            self._run.append({
                'id': report.nodeid,
                'outcome': report.outcome,
            })


    def write_results(self) -> None:
        if self._results_file:
            with self._results_file.open("w") as f:
                json.dump({
                    'collect': self._collect,
                    'run': self._run
                }, f)


@pytest.hookimpl
def pytest_addoption(parser: pytest.Parser, pluginmanager: pytest.PytestPluginManager) -> None:
    parser.addoption(MODULE_LIST_ARG, type=Path, help="Only collect modules in the given file")
    parser.addoption(TEST_LIST_ARG, type=Path, help="Only run tests whose node IDs are in the given file")
    parser.addoption(RESULTS_ARG, type=Path, help="Write test collection/run results to the given file")


@pytest.hookimpl
def pytest_configure(config: pytest.Config) -> None:
    if config.getoption(MODULE_LIST_ARG) or config.getoption(TEST_LIST_ARG) or \
       config.getoption(RESULTS_ARG):
        config._cleanslate_reduce_plugin = ReducePlugin(config)
        config.pluginmanager.register(config._cleanslate_reduce_plugin)


@pytest.hookimpl
def pytest_unconfigure(config: pytest.Config) -> None:
    if (plugin := getattr(config, "_cleanslate_reduce_plugin", None)):
        plugin.write_results()


class Results:
    """Facilitates access to test results file generated by ReducePlugin."""

    def __init__(self, results_file: Path):
        with results_file.open("r") as f:
            self._results = json.load(f)

        self._outcomes = None

    def get_outcome(self, nodeid: str) -> str:
        if self._outcomes is None:
            self._outcomes = {r['id']: r['outcome'] for r in self._results['collect'] + self._results['run']}

        return self._outcomes[nodeid]

    def get_modules(self) -> T.List[str]:
        return [r['id'] for r in self._results['collect']]

    def get_tests(self) -> T.List[str]:
        return [r['id'] for r in self._results['run']]

    def get_failed(self) -> T.Iterator[str]:
        return iter(o['id'] for o in self._results['collect'] + self._results['run'] if o['outcome'] == 'failed')

    def get_first_failed(self) -> T.Union[None, str]:
        return next(self.get_failed(), None)


def get_module(testid: str) -> str:
    return testid.split('::')[0]


def _is_module(testid: str) -> bool:
    return '::' not in testid


def get_function(testid: str) -> str:
    if '::' in testid:
        return testid.split('::')[1].split('[')[0]


def run_pytest(tests_path: Path, pytest_args=(), *,
               modules: T.List[Path] = None, tests: T.List[str] = None, trace: bool = False) -> dict:
    import tempfile
    import subprocess

    # Specifying dir='.' here works around weird failures running tests on MacOS
    with tempfile.TemporaryDirectory(dir='.') as tmpdir:
        tmpdir = Path(tmpdir)

        results = tmpdir / "results.json"

        if modules:
            modulelist = tmpdir / "modules.txt"
            modulelist.write_text('\n'.join(modules))

        if tests:
            testlist = tmpdir / "tests.txt"
            testlist.write_text('\n'.join(tests))

        command = [
            sys.executable, '-m', 'pytest', *PYTEST_ARGS, *pytest_args,
            RESULTS_ARG, results,
            *((MODULE_LIST_ARG, modulelist) if modules else ()),
            *((TEST_LIST_ARG, testlist) if tests else ()),
            tests_path
        ]

        if trace:
            print(f"Running {command}", flush=True)

        p = subprocess.run(command,
                           check=False, **({} if trace else {'stdout': subprocess.DEVNULL}))
        if p.returncode not in (pytest.ExitCode.OK, pytest.ExitCode.TESTS_FAILED,
                                pytest.ExitCode.INTERRUPTED, pytest.ExitCode.NO_TESTS_COLLECTED):
            p.check_returncode()

        return Results(results)


def _bisect_items(items: T.List[str], failing: str, fails: T.Callable[[T.List[str], str], bool],
                  *, bar: "tqdm") -> T.List[str]:
    assert failing not in items

    while len(items) > 1:
        middle = len(items) // 2

        bar.refresh() # for when using --trace
        bar.set_postfix({"remaining": len(items)})
        bar.update()

        if fails(items[:middle]+[failing]):
            items = items[:middle]
            continue

        if fails(items[middle:]+[failing]):
            items = items[middle:]
            continue

        # TODO could do the rest of delta debugging here
        break

    if len(items) == 1 and fails([failing]):
        items = []

    bar.refresh() # for when using --trace
    bar.set_postfix({"remaining": len(items)})
    bar.update()

    return items


def _reduce_tests(tests_path: Path, tests: T.List[str], failing_test: str, modules: T.List[str],
                  *, trace: bool = False, pytest_args: T.List[str] = ()) -> T.List[str]:
    def fails(test_set: T.List[str]):
        trial = run_pytest(tests_path, (*pytest_args, '--continue-on-collection-errors'),
                           tests=test_set, modules=modules, trace=trace)
        return trial.get_outcome(failing_test) == 'failed'

    module_set = {*modules}
    tests = [t for t in tests if t != failing_test and get_module(t) in module_set]
    if not tests:
        return tests

    steps=math.ceil(math.log(len(tests), 2))
    with tqdm.tqdm(desc="Trying to reduce tests.....", total=steps) as bar:
        return _bisect_items(tests, failing_test, fails, bar=bar)


def _reduce_modules(tests_path: Path, tests: T.List[str], failing_id: str,
                    modules: T.List[str], failing_module: str,
                    *, trace: bool = False, pytest_args: T.List[str] = ()) -> T.List[str]:

    def fails(module_set: T.List[str]):
        trial = run_pytest(tests_path, (*pytest_args, '--continue-on-collection-errors',),
                           tests=tests, modules=module_set, trace=trace)
        return trial.get_outcome(failing_id) == 'failed'

    modules = [m for m in modules if m != failing_module]
    if not modules:
        return modules

    steps = math.ceil(math.log(len(modules), 2))
    with tqdm.tqdm(desc="Trying to reduce modules...", total=steps) as bar:
        return _bisect_items(modules, failing_module, fails, bar=bar)


def reduce(*, tests_path: Path, results: Results = None, pytest_args: T.List[str] = (),
           trace: bool = False, **args) -> dict:
    if not results:
        print("Running tests...", flush=True)
        results = run_pytest(tests_path, (*pytest_args, '-x'), trace=trace)

    failed_id = results.get_first_failed()
    if failed_id is None:
        print("No tests failed!", flush=True)
        return {
            'failed': failed_id,
            'error': 'No tests failed',
        }

    failed_is_module = _is_module(failed_id)
    if failed_is_module:
        if trace: print()
        print(f"Module \"{failed_id}\"'s collection failed; trying it by itself...", flush=True)
        failed_module = failed_id
        tests = None
    else:
        if trace: print()
        print(f"Test \"{failed_id}\" failed; trying it by itself...", flush=True)
        failed_module = get_module(failed_id)
        tests = [failed_id]

    solo = run_pytest(tests_path, pytest_args, modules=[failed_module], tests=tests, trace=trace)
    if solo.get_outcome(failed_id) != 'passed':
        print("That also fails by itself!", flush=True)
        return {
            'failed': failed_id,
            'error': f'{"Module" if failed_is_module else "Test"} also fails by itself',
        }

    tests = results.get_tests()

    if trace: print()
    modules = _reduce_modules(tests_path, tests, failed_id, results.get_modules(), failed_module,
                              trace=trace, pytest_args=pytest_args)

    if not failed_is_module:
        if trace: print()
        tests = _reduce_tests(tests_path, tests, failed_id, [*modules, failed_module],
                              trace=trace, pytest_args=pytest_args)

        # TODO if tests != [], see if it's enough to disable just them

    if trace: print()
    print("Reduced failure set:")
    print(f"    modules: {modules}")
    print(f"    tests: {tests}")
    print(flush=True)

    return {
        'failed': failed_id,
        'modules': modules,
        'tests': tests,
    }


def _parse_args():
    import argparse

    bool_action = argparse.BooleanOptionalAction if sys.version_info[:2] >= (3,9) else "store_true"

    ap = argparse.ArgumentParser()
    ap.add_argument('--trace', default=False, action=bool_action, help='show pytest outputs, etc.')
    ap.add_argument('--save-to', type=Path, help='file where to save results (JSON)')
    ap.add_argument('--pytest-args', type=str, default='', help='extra arguments to pass to pytest')
    ap.add_argument('--version', action='version',
                    version=f"%(prog)s v{__version__} (Python {'.'.join(map(str, sys.version_info[:3]))})")
    ap.add_argument('tests_path', type=Path, help='tests file or directory')

    return ap.parse_args()


def main():
    args = _parse_args()
    args.pytest_args = args.pytest_args.split()

    results = reduce(**vars(args))
    if args.save_to:
        with args.save_to.open("w") as f:
            json.dump(results, f)

    return 1 if 'error' in results else 0


if __name__ == "__main__":
    sys.exit(main())



import pytest
import subprocess
import typing as T
import sys
from test_cleanslate import seq2p, tests_dir, make_polluted_suite, FAILURES
import json
from textwrap import dedent
from pathlib import Path

import pytest_cleanslate.reduce as reduce

from pytest_cleanslate.reduce import MODULE_LIST_ARG, TEST_LIST_ARG, RESULTS_ARG, Results, \
            get_module, get_function


def test_get_module():
    assert 'test.py' == get_module('test.py')
    assert 'test.py' == get_module('test.py::test_foo')
    assert 'test.py' == get_module('test.py::test_foo[1]')


def test_get_function():
    assert None == get_function('test.py')
    assert 'test_foo' == get_function('test.py::test_foo')
    assert 'test_foo' == get_function('test.py::test_foo[1]')


def test_run_pytest_collect_failure(tests_dir):
    test1 = seq2p(tests_dir, 1)
    test1.write_text(dedent("""\
        def test_one():
            assert True
        """))

    test2 = seq2p(tests_dir, 2)
    test2.write_text(dedent("""\
        blargh this ain't python
        """))

    r = reduce.run_pytest(tests_dir)

    assert r.get_outcome(f"{test1}") == 'passed'
    assert r.get_outcome(f"{test2}") == 'failed'


def test_run_pytest_module_list(tests_dir):
    test1 = seq2p(tests_dir, 1)
    test1.write_text(dedent("""\
        def test_one():
            assert True

        def test_two():
            assert False
        """))

    test2 = seq2p(tests_dir, 2)
    test2.write_text(dedent("""\
        blargh this ain't python
        """))

    test3 = seq2p(tests_dir, 3)
    test3.write_text(dedent("""\
        def test_nothing():
            pass
        """))

    r = reduce.run_pytest(tests_dir, modules=[str(test1), str(test3)])
    ids = r.get_tests()
    # the order should be as executed
    assert ids.index(f"{test1}::test_one") < ids.index(f"{test1}::test_two")

    assert r.get_outcome(f"{test1}::test_one") == 'passed'
    assert r.get_outcome(f"{test1}::test_two") == 'failed'
    assert r.get_outcome(f"{test3}::test_nothing") == 'passed'


def test_run_pytest_test_list(tests_dir):
    test1 = seq2p(tests_dir, 1)
    test1.write_text(dedent("""\
        def test_one():
            assert False

        def test_two():
            assert True
        """))

    test2 = seq2p(tests_dir, 2)
    test2.write_text(dedent("""\
        def test_one():
            assert True 
        """))

    tests = tests_dir.parent / "tests.txt"
    tests.write_text(dedent(f"""\
        {test1}::test_two
        {test2}::test_one
        """))

    r = reduce.run_pytest(tests_dir, tests=[f"{test1}::test_two", f"{test2}::test_one"])
    assert r.get_outcome(f"{test1}") == 'passed'
    assert r.get_outcome(f"{test2}") == 'passed'
    assert f"{test1}::test_one" not in r.get_tests()
    assert r.get_outcome(f"{test1}::test_two") == 'passed'
    assert r.get_outcome(f"{test2}::test_one") == 'passed'


def cli_reduce(*, tests_path: Path, pytest_args: T.List[str] = (), trace: bool = False, **args) -> dict:
    reduction_file = tests_path.parent / "reduction.json"

    p = subprocess.run([sys.executable, '-m', 'pytest_cleanslate.reduce',
                        '--save-to', reduction_file,
                        *((f"--pytest-args={' '.join(pytest_args)}",) if pytest_args else ()),
                        *(("--trace",) if trace else ()),
                        *(f"--{name}={value}" for name, value in args.items()),
                        tests_path], check=False)

    with reduction_file.open("r") as f:
        reduction = json.load(f)

    assert 'error' not in reduction or p.returncode == 1
    return reduction


@pytest.mark.parametrize("n_tests", [0, 1, 10])
@pytest.mark.parametrize("r", [reduce.reduce, cli_reduce])
def test_reduce_nothing_fails(tests_dir, n_tests, r):
    for seq in range(n_tests):
        seq2p(tests_dir, seq).write_text('def test_foo(): pass')

    reduction = r(tests_path=tests_dir, trace=True)

    assert reduction['failed'] == None
    assert 'error' in reduction


@pytest.mark.parametrize("r", [reduce.reduce, cli_reduce])
def test_reduce_test_fails_by_itself(tests_dir, r):
    for seq in range(10):
        seq2p(tests_dir, seq).write_text('def test_foo(): pass')

    failing = seq2p(tests_dir, 3)
    failing.write_text('def test_foo(): assert False')

    reduction = r(tests_path=tests_dir, trace=True)

    assert reduction['failed'] == f"{failing}::test_foo"
    assert 'error' in reduction


@pytest.mark.parametrize("pollute_in_collect, fail_collect", [[False, False], [True, False], [True, True]])
@pytest.mark.parametrize("r", [reduce.reduce, cli_reduce])
def test_reduce(tests_dir, pollute_in_collect, fail_collect, r):
    failing, polluter, tests = make_polluted_suite(tests_dir, fail_collect=fail_collect,
                                                   pollute_in_collect=pollute_in_collect)

    reduction_file = tests_dir.parent / "reduction.json"

    reduction = r(tests_path=tests_dir, trace=True)

    assert reduction['failed'] == failing
    assert reduction['modules'] == [get_module(polluter)]
# this would be more precise... is it the test or the module?
#    assert reduction['modules'] == ([get_module(polluter)] if pollute_in_collect else [])
    assert reduction['tests'] == ([] if pollute_in_collect else [polluter])


@pytest.mark.parametrize("pollute_in_collect, fail_collect", [[False, False], [True, False], [True, True]])
@pytest.mark.parametrize("r", [reduce.reduce, cli_reduce])
def test_reduce_pytest_args(tests_dir, pollute_in_collect, fail_collect, r):
    failing, polluter, tests = make_polluted_suite(tests_dir, fail_collect=fail_collect,
                               pollute_in_collect=pollute_in_collect)

    (tests_dir / "conftest.py").write_text(dedent("""\
        if read, this breaks everything
        """))

    reduction = r(tests_path=tests_dir, trace=True, pytest_args=['--noconftest'])

    assert reduction['failed'] == failing
    assert reduction['modules'] == [get_module(polluter)]
# this would be more precise... is it the test or the module?
#    assert reduction['modules'] == ([get_module(polluter)] if pollute_in_collect else [])
    assert reduction['tests'] == ([] if pollute_in_collect else [polluter])


@pytest.mark.parametrize("r", [reduce.reduce, cli_reduce])
def test_reduce_other_collection_fails(tests_dir, r):
    """Tests that we use --continue-on-collection-errors"""
    failing, polluter, tests = make_polluted_suite(tests_dir, pollute_in_collect=True, fail_collect=False,
                                                   polluter_seq = 3, failing_seq = 8)

    seq2p(tests_dir, 0).write_text(dedent("""\
        import sys
        sys.needs_this = True

        def test_nothing():
            assert True
        """))

    # this one fails collection pytest runs set to ignore the one above
    seq2p(tests_dir, 2).write_text(dedent("""\
        import sys

        if not hasattr(sys, 'needs_this'):
            raise RuntimeError('argh')

        def test_nothing():
            assert True
        """))

    reduction = r(tests_path=tests_dir, trace=True)

    assert reduction['failed'] == failing
    assert reduction['modules'] == [get_module(polluter)]
    assert reduction['tests'] == []


@pytest.mark.parametrize("r", [reduce.reduce, cli_reduce])
def test_reduce_polluter_test_in_single_module(tests_dir, r):
    test = seq2p(tests_dir, 0)
    test.write_text(dedent("""\
        import sys

        def test_polluter():
            sys.needs_this = True
            assert True

        def test_nothing():
            assert True

        def test_failing():
            assert not hasattr(sys, 'needs_this')
        """))

    reduction = r(tests_path=tests_dir, trace=True)

    assert reduction['failed'] == f"{str(test)}::test_failing"
    assert reduction['modules'] == []
    assert reduction['tests'] == [f"{str(test)}::test_polluter"]



import pytest
import sys
import subprocess
from pathlib import Path
import json
import random
from textwrap import dedent

pytestmark = pytest.mark.skipif(sys.platform == 'win32', reason='Unix-only')


def seq2p(tests_dir, seq):
    return tests_dir / f"test_{seq:02}.py"


FAILURES = {
    'assert': 'assert False',
    'exception': 'raise RuntimeError("test")',
    'kill': 'os.kill(os.getpid(), 9)',
    'exit': 'pytest.exit("goodbye")',
    'interrupt': 'raise KeyboardInterrupt()'
}

def make_polluted_suite(tests_dir: Path, *, pollute_in_collect: bool = True, fail_collect: bool = False,
                        fail_kind: str = 'assert', polluter_seq: int = None, failing_seq: int = None):
    """in a suite with 10 tests, 'polluter' doesn't fail, but causes 'failing' to fail."""
    # note the polluter must run before the failing test
    N_TESTS = 10

    assert (polluter_seq is None and failing_seq is None) or (failing_seq > polluter_seq)

    if polluter_seq is None:
        polluter_seq = random.choice(range(N_TESTS-1))

    polluter = seq2p(tests_dir, polluter_seq)
    if pollute_in_collect:
        polluter.write_text(dedent("""\
            import sys
            sys.foobar=True

            def test_bar():
                assert True
            """))
    else:
        polluter.write_text(dedent("""\
            import sys

            def test_polluter():
                sys.foobar=True
                assert True

            def test_bar():
                assert True
            """))
        polluter = f"{polluter}::test_polluter"

    if failing_seq is None:
        failing_seq = random.choice(range(polluter_seq+1, N_TESTS))

    failing = seq2p(tests_dir, failing_seq)
    if fail_collect:
        failing.write_text(dedent(f"""\
            import sys
            import os
            import pytest

            if getattr(sys, 'foobar', False):
                {FAILURES[fail_kind]}

            def test_ok():
                assert True
            """))
    else:
        failing.write_text(dedent(f"""\
            import sys
            import os
            import pytest

            def test_failing():
                if getattr(sys, 'foobar', False):
                    {FAILURES[fail_kind]}

            def test_ok():
                assert True
            """))
        failing = f"{failing}::test_failing"

    tests = [seq for seq in range(N_TESTS) if seq not in (polluter_seq, failing_seq)]
    for seq in tests:
        seq2p(tests_dir, seq).write_text('def test_foo(): pass')

    return str(failing), str(polluter), tests


@pytest.fixture
def tests_dir(tmp_path, monkeypatch):
    monkeypatch.chdir(tmp_path)
    tests_dir = Path('tests')
    tests_dir.mkdir()
    yield tests_dir


@pytest.mark.parametrize("pollute_in_collect, fail_collect", [[False, False], [True, False], [True, True]])
@pytest.mark.parametrize("fail_kind", list(FAILURES.keys() - {'kill'}))
def test_check_suite_fails(tests_dir, pollute_in_collect, fail_collect, fail_kind):
    make_polluted_suite(tests_dir, pollute_in_collect=pollute_in_collect,
                        fail_collect=fail_collect, fail_kind=fail_kind)

    p = subprocess.run([sys.executable, '-m', 'pytest', tests_dir], check=False)
    if fail_collect or fail_kind in ('interrupt', 'exit'):
        assert p.returncode == pytest.ExitCode.INTERRUPTED
    else:
        assert p.returncode == pytest.ExitCode.TESTS_FAILED

    p = subprocess.run([sys.executable, '-m', 'pytest',
                        '--continue-on-collection-errors', tests_dir], check=False)
    # pytest versions disagree on what the exit code should be here; more current
    # ones seem to side with INTERRUPTED for 'interrupt' and 'exit'
    assert p.returncode in (pytest.ExitCode.INTERRUPTED, pytest.ExitCode.TESTS_FAILED)

@pytest.mark.parametrize("plugin", ['asyncio', 'no:asyncio'])
@pytest.mark.parametrize("pollute_in_collect, fail_collect", [[False, False], [True, False], [True, True]])
@pytest.mark.parametrize("fail_kind", list(FAILURES.keys()))
def test_isolate_polluted(tests_dir, pollute_in_collect, fail_collect, fail_kind, plugin):
    make_polluted_suite(tests_dir, pollute_in_collect=pollute_in_collect,
                        fail_collect=fail_collect, fail_kind=fail_kind)

    p = subprocess.run([sys.executable, '-m', 'pytest', '-p', plugin, '--cleanslate', tests_dir], check=False)
    assert p.returncode == pytest.ExitCode.OK


@pytest.mark.parametrize("fail_kind", list(FAILURES.keys()))
def test_pytest_discover_tests(tests_dir, fail_kind):
    make_polluted_suite(tests_dir, fail_collect=False, fail_kind=fail_kind)

    p = subprocess.run([sys.executable, '-m', 'pytest', '--cleanslate'], check=False) # no tests_dir here
    assert p.returncode == pytest.ExitCode.OK


@pytest.mark.parametrize("pollute_in_collect, fail_collect", [[False, False], [True, False], [True, True]])
@pytest.mark.parametrize("fail_kind", list(FAILURES.keys() - {'kill'}))
def test_unconditionally_failing_test(tests_dir, pollute_in_collect, fail_collect, fail_kind):
    _, _, tests = make_polluted_suite(tests_dir, pollute_in_collect=pollute_in_collect,
                                      fail_collect=fail_collect, fail_kind=fail_kind)

    # _unconditionally_ failing test
    failing = seq2p(tests_dir, random.choice(tests))
    failing.write_text(f"""\
import sys
import os
import pytest

def failure():
    {FAILURES[fail_kind]}

def test_foo():
    failure()

{'test_foo()' if fail_collect else ''}
""")

    p = subprocess.run([sys.executable, '-m', 'pytest', '--cleanslate', tests_dir], check=False,
                       capture_output=True)
    print(failing.read_text())
    if fail_collect or fail_kind in ('interrupt', 'exit'):
        assert p.returncode == pytest.ExitCode.INTERRUPTED
    else:
        assert p.returncode == pytest.ExitCode.TESTS_FAILED
    # pytest-forked error message that shouldn't be used unless the process was truly killed
    assert 'CRASHED with signal 0' not in str(p.stdout, 'utf-8')


    p = subprocess.run([sys.executable, '-m', 'pytest',
                        '--continue-on-collection-errors', tests_dir],
                       check=False, capture_output=True)
    if fail_kind == 'interrupt' or (fail_kind == 'exit' and not fail_collect):
        assert p.returncode == pytest.ExitCode.INTERRUPTED
    else:
        assert p.returncode == pytest.ExitCode.TESTS_FAILED
    assert 'CRASHED with signal 0' not in str(p.stdout, 'utf-8')


def test_collect_failure(tests_dir):
    # _unconditionally_ failing test
    failing = seq2p(tests_dir, 1)
    failing.write_text(f"""\
import sys
import os
import pytest

def test_foo():
    failure()

assert False
""")

    p = subprocess.run([sys.executable, '-m', 'pytest', '--cleanslate', tests_dir], check=False,
                       capture_output=True)
    print(str(p.stdout, 'utf-8'))
    assert p.returncode == pytest.ExitCode.INTERRUPTED

    # pytest-forked error message that shouldn't be used unless the process was truly killed
    assert 'CRASHED with signal 0' not in str(p.stdout, 'utf-8')


def test_isolate_module_yields_collector(tests_dir):
    # A pytest.Collector.collect()'s return value may include not only pytest.Item,
    # but also pytest.Collector.
    #
    # Here we test for this by including a class within the test module:
    # when the module is being collected, pytest.Module.collect() will include
    # a pytest.Class collector to actually collect items from within the class.

    test = seq2p(tests_dir, 1)
    test.write_text("""\
class TestClass:
    def test_foo(self):
        assert True
""")

    p = subprocess.run([sys.executable, '-m', 'pytest', '--cleanslate', tests_dir], check=False)
    assert p.returncode == pytest.ExitCode.OK


# If asyncio is missing/disabled, the test may show as skipped; we detect it here
# with the 'fail=True' version of the test.
@pytest.mark.parametrize("fail", [False, True])
def test_asyncio(tests_dir, fail):
    test = seq2p(tests_dir, 1)
    test.write_text(f"""\
import pytest
import asyncio

async def foo(s):
    return s

@pytest.mark.asyncio
async def test_asyncio():
    assert "bar" {'!=' if fail else '=='} await foo("bar")
""")

    p = subprocess.run([sys.executable, '-m', 'pytest', '--cleanslate', tests_dir], check=False)
    if fail:
        assert p.returncode == pytest.ExitCode.TESTS_FAILED
    else:
        assert p.returncode == pytest.ExitCode.OK


@pytest.mark.parametrize("tf", ['test_one', 'test_two'])
def test_mark(tests_dir, tf):
    # the built-in "mark" plugin implements '-k' and '-m'

    test = seq2p(tests_dir, 1)
    test.write_text("""\
def test_one():
    assert True

def test_two():
    assert False
""")

    p = subprocess.run([sys.executable, '-m', 'pytest', '--cleanslate', '-k', tf, tests_dir], check=False)
    if tf == 'test_two':
        assert p.returncode == pytest.ExitCode.TESTS_FAILED
    else:
        assert p.returncode == pytest.ExitCode.OK


def test_exitfirst(tests_dir):
    test = seq2p(tests_dir, 1)
    test.write_text("""\
from pathlib import Path

def test_one():
    assert False

def test_two():
    Path('litmus.txt').touch()
""")

    p = subprocess.run([sys.executable, '-m', 'pytest', '--cleanslate', '--exitfirst', '-s', tests_dir], check=False,
                       capture_output=True)
    assert p.returncode == pytest.ExitCode.TESTS_FAILED
    assert not Path('litmus.txt').exists()
    assert 'CRASHED' not in str(p.stdout, 'utf-8')


def test_shouldstop(tests_dir):
    test = seq2p(tests_dir, 1)
    test.write_text("""\
import pytest
from pathlib import Path

def test_one():
    assert False

def test_two():
    Path('litmus.txt').touch()
""")

    # --stepwise sets session.shouldstop upon a test failure.
    p = subprocess.run([sys.executable, '-m', 'pytest', '--cleanslate', '--stepwise', '-s', tests_dir], check=False,
                       capture_output=True)
    assert p.returncode == pytest.ExitCode.INTERRUPTED
    assert not Path('litmus.txt').exists()
    assert 'CRASHED' not in str(p.stdout, 'utf-8')


def test_polluter_test_in_single_module(tests_dir):
    test = seq2p(tests_dir, 0)
    test.write_text(dedent("""\
        import sys

        def test_polluter():
            sys.needs_this = True
            assert True

        def test_nothing():
            assert True

        def test_failing():
            assert not hasattr(sys, 'needs_this')
        """))

    p = subprocess.run([sys.executable, '-m', 'pytest', '--cleanslate', tests_dir], check=False)
    assert p.returncode == pytest.ExitCode.OK


def test_closes_stdout_stderr(tests_dir):
    test = seq2p(tests_dir, 0)
    test.write_text(dedent("""\
        import os

        os.close(1)
        os.close(2)
        """))

    p = subprocess.run([sys.executable, '-m', 'pytest', '--cleanslate', tests_dir], check=False)
    assert p.returncode == pytest.ExitCode.OK



import pytest
from pathlib import Path


# py.process.ForkedFunc does os.close(1) and os.close(2) just before
# the exit of the child process... but if the function called also
# closes those files, an OSError results.  Here we ignore those...
# Really we need to move away from 'py'...
import os
import errno
class IgnoreOsCloseErrors:
    def __enter__(self):
        self.original_os_close = os.close

        def ignoring_close(fd):
            try:
                self.original_os_close(fd)
            except OSError as e:
                if fd not in (1, 2) or e.errno != errno.EBADF:
                    raise

        os.close = ignoring_close

    def __exit__(self, exc_type, exc_value, traceback):
        os.close = self.original_os_close


class CleanSlateItem(pytest.Item):
    """Item that stands for a Module until it can be collected from its forked subprocess"""
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def runtest(self):
        raise RuntimeError("This should never execute")

    def collect_and_run(self):
        # adapted from pytest-forked
        import pickle
        import _pytest
        import pytest_forked as ptf # FIXME pytest-forked is unmaintained
        import py                   # FIXME py is maintenance only

        def runforked():
            # Use 'parent' as it would have been in pytest_pycollect_makemodule, so that our
            # nodes aren't included in the chain, as they might confuse other plugins (such as 'mark')
            module = pytest.Module.from_parent(parent=self.parent.parent, path=self.path)

            def collect_items(collector):
                for it in collector.collect():
                    if isinstance(it, pytest.Collector):
                        yield from collect_items(it)
                    else:
                        yield it

            try:
                self.session.items = list(collect_items(module))
            except BaseException:
                excinfo = pytest.ExceptionInfo.from_current()
                return pickle.dumps([pytest.CollectReport(
                            nodeid=self.nodeid,
                            outcome='failed',
                            result=None,
                            longrepr=self._repr_failure_py(excinfo, "short"))
                ])

            pm = self.config.pluginmanager
            caller = pm.subset_hook_caller('pytest_collection_modifyitems', remove_plugins=[self.parent.plugin])
            caller(session=self.session, config=self.config, items=self.session.items)

            reports = list()
            class ReportSaver:
                @pytest.hookimpl
                def pytest_runtest_logreport(self, report):
                    reports.append(report)

            pm.register(ReportSaver())

            try:
                self.ihook.pytest_runtestloop(session=self.session)
            except (pytest.Session.Interrupted, pytest.Session.Failed):
                pass
            except BaseException as e:
                return pickle.dumps(e)

            return pickle.dumps(reports)

        with IgnoreOsCloseErrors():
            ff = py.process.ForkedFunc(runforked)
        result = ff.waitfinish()

        if result.retval is None:
            return [ptf.report_process_crash(self, result)]

        retval = pickle.loads(result.retval)
        if isinstance(retval, BaseException):
            raise retval

        return retval


class CleanSlateCollector(pytest.File, pytest.Collector):
    def __init__(self, *, plugin, **kwargs):
        super().__init__(**kwargs)
        self.plugin = plugin

    def collect(self):
        yield CleanSlateItem.from_parent(parent=self, name=self.name)


def run_item_forked(item):
    import _pytest.runner
    import pytest_forked as ptf # FIXME pytest-forked is unmaintained
    import py                   # FIXME py is maintenance only
    import pickle

    def runforked():
        try:
            return pickle.dumps(_pytest.runner.runtestprotocol(item, log=False))
        except BaseException as e:
            return pickle.dumps(e)

    ff = py.process.ForkedFunc(runforked)
    result = ff.waitfinish()

    if result.retval is None:
        return [ptf.report_process_crash(item, result)]

    retval = pickle.loads(result.retval)

    if isinstance(retval, BaseException):
        raise retval

    return retval


class CleanSlatePlugin:
    """Pytest plugin to isolate test collection, so that if a test's collection pollutes the in-memory
       state, it doesn't affect the execution of other tests."""


    @pytest.hookimpl(tryfirst=True)
    def pytest_pycollect_makemodule(self, module_path: Path, parent):
        return CleanSlateCollector.from_parent(parent, path=module_path, plugin=self)


    @pytest.hookimpl(tryfirst=True)
    def pytest_runtest_protocol(self, item: pytest.Item, nextitem: pytest.Item):
        ihook = item.ihook
        ihook.pytest_runtest_logstart(nodeid=item.nodeid, location=item.location)
        if isinstance(item, CleanSlateItem):
            reports = item.collect_and_run()
        else:
            # note any side effects, such as setting session.shouldstop, are lost...
            reports = run_item_forked(item)

        if (reports and isinstance(reports[0], pytest.CollectReport) and reports[0].outcome == 'failed'
            and not item.config.option.continue_on_collection_errors):
            item.session.shouldstop = 'collection error'

        for rep in reports:
            ihook.pytest_runtest_logreport(report=rep)

        ihook.pytest_runtest_logfinish(nodeid=item.nodeid, location=item.location)
        return True


    @pytest.hookimpl(tryfirst=True, hookwrapper=True)
    def pytest_collection_modifyitems(self, session, config, items):
        # Since we're deferring collection to CleanSlateItem, we won't have
        # functions and other relevant items in 'items', possibly leading to
        # our CleanSlateItems being deselected.
        # There doesn't seem to be a way to prevent other plugins from modifying
        # the list, so we save it, let them run, and restore it.
        initial_items = list(items) # TODO save them using pytest_deselected() instead?
        yield
        items[:] = initial_items


def pytest_addoption(parser: pytest.Parser, pluginmanager: pytest.PytestPluginManager) -> None:
    g = parser.getgroup('cleanslate')
    g.addoption("--cleanslate", action="store_true",
                help="Isolate test module collection and test execution using sys.fork()")


def pytest_configure(config: pytest.Config) -> None:
    if config.getoption("--cleanslate"):
        config.pluginmanager.register(CleanSlatePlugin(), "cleanslate_plugin")
